==> concierge-1.0.0-sources.jar: com/flybits/concierge/viewmodels/CategoryViewModel.kt <==
package com.flybits.concierge.viewmodels

import android.app.Activity
import android.arch.lifecycle.LiveData
import android.arch.lifecycle.ViewModel
import android.arch.paging.DataSource
import android.arch.paging.LivePagedListBuilder
import android.arch.paging.PagedList
import com.flybits.android.kernel.ContentAnalytics

import com.flybits.android.kernel.models.Content
import com.flybits.android.kernel.models.results.ContentResult
import com.flybits.android.kernel.utilities.ContentParameters
import com.flybits.commons.library.api.FlybitsManager
import com.flybits.commons.library.api.results.callbacks.ConnectionResultCallback
import com.flybits.commons.library.api.results.callbacks.ObjectResultCallback
import com.flybits.commons.library.api.results.callbacks.PagedResultCallback
import com.flybits.commons.library.exceptions.APIUsageExceededException
import com.flybits.commons.library.exceptions.FlybitsException
import com.flybits.commons.library.exceptions.NotConnectedException
import com.flybits.commons.library.logging.Logger
import com.flybits.concierge.*
import com.flybits.concierge.models.BaseTemplate
import com.flybits.concierge.repository.BaseTemplateDataSource
import com.flybits.concierge.repository.ContentRepository
import com.flybits.concierge.repository.ModelConverter
import java.util.*
import java.util.concurrent.*

import kotlin.concurrent.schedule

class CategoryViewModel(private val category: String
                        , private val flybitsConcierge: FlybitsConcierge
                        , private val contentRepository: ContentRepository
                        , private val contentAnalytics: ContentAnalytics
                        , private val resourceProvider: ResourceProvider
                        , private val modelConverter: ModelConverter) : ViewModel(), PagedResultCallback<Content> {

    data class ScrollEvent(val visibleItems: List<BaseTemplate>)

    interface FeedErrorDisplayer {
        fun onError(err: String)
    }

    companion object {
        const val PAGE_SIZE = 20
        const val PRE_FETCH_DISTANCE = 10
        const val INITIAL_LOAD_SIZE = PAGE_SIZE
        const val VIEWING_THRESHOLD = 250L
    }

    private var contentResult: ContentResult? = null
    private var onRefreshFinishedCallback: (() -> Unit)? = null
    private var feedErrorDisplayer: FeedErrorDisplayer? = null
    private var loadedAll: Boolean = false
    private val currentTaskList: ConcurrentHashMap<String, TimerTask> = ConcurrentHashMap()
    private val scrollEventQueue: BlockingQueue<ScrollEvent> = LinkedBlockingQueue()
    private var contentDataSource: DataSource<Int, BaseTemplate>? = null

    val feedContent: LiveData<PagedList<BaseTemplate?>>
        get() {
            loadedAll = false
            val config = PagedList.Config.Builder()
                    .setInitialLoadSizeHint(INITIAL_LOAD_SIZE)
                    .setPageSize(PAGE_SIZE)
                    .setEnablePlaceholders(true)
                    .setPrefetchDistance(PRE_FETCH_DISTANCE)
                    .build()

            val typeList = ArrayList<String>()
            for (viewProvider in flybitsConcierge.flybitsViewProviders) {
                typeList.add(viewProvider.getContentType())
            }
            Logger.d("Category: $category, returning LivePagedList")
            val dataSourceFactory = BaseTemplateDataSource.getFactory(category, contentRepository, modelConverter) {
                contentDataSource = it
            }
            return LivePagedListBuilder(dataSourceFactory, config)
                    .setBoundaryCallback(object : PagedList.BoundaryCallback<BaseTemplate?>() {

                override fun onZeroItemsLoaded() {
                    Logger.d("onZeroItemsLoaded()")
                    requestContent()
                }

                override fun onItemAtEndLoaded(itemAtEnd: BaseTemplate) {
                    Logger.d("Category: $category, onItemAtEndLoaded() ")
                    if (!loadedAll) {
                        contentResult = contentResult?.getMore(this@CategoryViewModel) as ContentResult?
                    }
                }
            }).build()
        }

    init {

        setupScrollHandler()

        loadedAll = false
    }

    fun refresh(refreshResultListener: () -> Unit = {}) {
        loadedAll = false
        contentResult = null
        requestContent()
        this.onRefreshFinishedCallback = refreshResultListener
    }

    private fun requestContent() {
        if (contentResult == null) {
            val paramsBuilder = ContentParameters.Builder()
                    .setPaging(PAGE_SIZE.toLong(), 0)
                    .setCaching(category, INITIAL_LOAD_SIZE)

            //Ignore labels for home tab, show all
            if (!category.equals(resourceProvider.getString(R.string.flybits_con_default_tab_name), ignoreCase = true)){
                paramsBuilder.setLabels(category)
            }

            val params = paramsBuilder.build()
            contentResult = contentRepository.getContent(params,this)
        }
    }

    fun setFeedErrorDisplayer(feedErrorDisplayer: FeedErrorDisplayer) {
        this.feedErrorDisplayer = feedErrorDisplayer
    }

    override fun onSuccess(items: ArrayList<Content>) {
        Logger.d("Category: $category, onSuccess() items count: " + items.size)
        onRefreshFinishedCallback?.invoke()
        onRefreshFinishedCallback = null
        contentDataSource?.invalidate()
    }

    override fun onException(exception: FlybitsException) {
        Logger.exception(javaClass.simpleName, exception)

        if (exception is NotConnectedException || exception is APIUsageExceededException) {
            flybitsConcierge.unauthenticateWithoutLogout(exception)
        }
        onRefreshFinishedCallback?.invoke()
        onRefreshFinishedCallback = null

        if (feedErrorDisplayer != null) {
            feedErrorDisplayer!!.onError(resourceProvider.getString(R.string.flybits_con_generic_error))
        }

    }

    override fun onLoadedAllItems() {
        loadedAll = true
    }

    fun connectAndDownLoad(activity: Activity, metaDataId: String, callback: ObjectResultCallback<Content>?) {
        FlybitsManager.isConnected(activity, true, object : ConnectionResultCallback {
            override fun onConnected() {
                Content.getById(activity, metaDataId, callback)
            }

            override fun notConnected() {

            }

            override fun onException(exception: FlybitsException) {
                callback?.onException(exception)
            }
        })

    }

    //these scroll events need to be queued to prevent a concurrent modification exception
    fun onScroll(visibleItems: List<BaseTemplate>){
        scrollEventQueue.add(ScrollEvent(visibleItems))
    }

    private fun setupScrollHandler(){
        Executors.newSingleThreadExecutor().execute {
            while (true){
                val visibleItems = scrollEventQueue.take().visibleItems

                val visibleContentList = visibleItems.asSequence().map { it.content }.filter { it != null }.toList()

                //Cancel then remove tasks which disappeared from the screen after scrolling
                currentTaskList.keys.asSequence()
                        .filter {
                            //Filter out all the content that can be found in the visible list
                            visibleContentList.find { item -> item.id == it } == null
                        }.forEach {
                            //Cancel the timer for the view that disappeared
                            currentTaskList[it]?.cancel()
                            currentTaskList.remove(it)
                        }

                //Schedule the view actions for views which just appeared on the screen after scrolling
                visibleContentList.forEach {
                    //This timer will be canceled if this view disappears before the 250ms timer is done
                    if (currentTaskList[it.id] == null){
                        currentTaskList[it.id] = Timer().schedule(VIEWING_THRESHOLD){
                            contentAnalytics.trackViewed(it, System.currentTimeMillis())
                            currentTaskList.remove(it.id)
                        }
                    }
                }
            }
        }
    }
}

==> concierge-1.0.1-sources.jar: com/flybits/concierge/viewmodels/CategoryViewModel.kt <==
package com.flybits.concierge.viewmodels

import android.app.Activity
import android.arch.lifecycle.LiveData
import android.arch.lifecycle.ViewModel
import android.arch.paging.DataSource
import android.arch.paging.LivePagedListBuilder
import android.arch.paging.PagedList
import com.flybits.android.kernel.ContentAnalytics

import com.flybits.android.kernel.models.Content
import com.flybits.android.kernel.models.results.ContentResult
import com.flybits.android.kernel.utilities.ContentParameters
import com.flybits.commons.library.api.FlybitsManager
import com.flybits.commons.library.api.results.callbacks.ConnectionResultCallback
import com.flybits.commons.library.api.results.callbacks.ObjectResultCallback
import com.flybits.commons.library.api.results.callbacks.PagedResultCallback
import com.flybits.commons.library.exceptions.APIUsageExceededException
import com.flybits.commons.library.exceptions.FlybitsException
import com.flybits.commons.library.exceptions.NotConnectedException
import com.flybits.commons.library.logging.Logger
import com.flybits.concierge.*
import com.flybits.concierge.models.BaseTemplate
import com.flybits.concierge.repository.BaseTemplateDataSource
import com.flybits.concierge.repository.ContentRepository
import com.flybits.concierge.repository.ModelConverter
import java.util.*
import java.util.concurrent.*

import kotlin.concurrent.schedule

class CategoryViewModel(private val category: String
                        , private val flybitsConcierge: FlybitsConcierge
                        , private val contentRepository: ContentRepository
                        , private val contentAnalytics: ContentAnalytics
                        , private val resourceProvider: ResourceProvider
                        , private val modelConverter: ModelConverter) : ViewModel(), PagedResultCallback<Content> {

    data class ScrollEvent(val visibleItems: List<BaseTemplate>)

    interface FeedErrorDisplayer {
        fun onError(err: String)
    }

    companion object {
        const val PAGE_SIZE = 20
        const val PRE_FETCH_DISTANCE = 10
        const val INITIAL_LOAD_SIZE = PAGE_SIZE
        const val VIEWING_THRESHOLD = 250L
    }

    private var contentResult: ContentResult? = null
    private var onRefreshFinishedCallback: (() -> Unit)? = null
    private var feedErrorDisplayer: FeedErrorDisplayer? = null
    private var loadedAll: Boolean = false
    private val currentTaskList: ConcurrentHashMap<String, TimerTask> = ConcurrentHashMap()
    private val scrollEventQueue: BlockingQueue<ScrollEvent> = LinkedBlockingQueue()
    private var contentDataSource: DataSource<Int, BaseTemplate>? = null
    private var gotLocal = false
    private var loadedContent = false

    val feedContent: LiveData<PagedList<BaseTemplate?>>
        get() {
            loadedAll = false
            val config = PagedList.Config.Builder()
                    .setInitialLoadSizeHint(INITIAL_LOAD_SIZE)
                    .setPageSize(PAGE_SIZE)
                    .setEnablePlaceholders(true)
                    .setPrefetchDistance(PRE_FETCH_DISTANCE)
                    .build()

            val typeList = ArrayList<String>()
            for (viewProvider in flybitsConcierge.flybitsViewProviders) {
                typeList.add(viewProvider.getContentType())
            }
            Logger.d("Category: $category, returning LivePagedList")
            val dataSourceFactory = BaseTemplateDataSource.getFactory(category, contentRepository, modelConverter) {
                contentDataSource = it
            }

            val livePagedList =  LivePagedListBuilder(dataSourceFactory, config)
                    .setBoundaryCallback(object : PagedList.BoundaryCallback<BaseTemplate?>() {

                override fun onZeroItemsLoaded() {
                    Logger.d("onZeroItemsLoaded()")
                    requestContent()
                }

                override fun onItemAtEndLoaded(itemAtEnd: BaseTemplate) {
                    Logger.d("Category: $category, onItemAtEndLoaded() ")
                    if (!loadedAll) {
                        contentResult = contentResult?.getMore(this@CategoryViewModel) as ContentResult?
                    }
                }
            }).build()

            /* We need to invalidate the datasource here in case content has been loaded remotely but
            *  the data source was not invalidated in the onSuccess() method call because it was waiting
            *  for the local data to be retrieved first. If we do not do this then remote data won't
            *  come back to the UI
            *
            *  The if statement should only execute once, when the feed is first loaded*/
            livePagedList.observeForever {
                Logger.d("Category: $category, got local data, loadedContent = $loadedAll gotLocal = $gotLocal")
                if (loadedContent && !gotLocal){
                    contentDataSource?.invalidate()
                }
                gotLocal = true
            }

            return livePagedList
        }

    init {

        setupScrollHandler()

        loadedAll = false
    }

    fun refresh(refreshResultListener: () -> Unit = {}) {
        loadedAll = false
        contentResult = null
        requestContent()
        this.onRefreshFinishedCallback = refreshResultListener
    }

    private fun requestContent() {
        if (contentResult == null) {
            val paramsBuilder = ContentParameters.Builder()
                    .setPaging(PAGE_SIZE.toLong(), 0)
                    .setCaching(category, INITIAL_LOAD_SIZE)

            //Ignore labels for home tab, show all
            if (!category.equals(resourceProvider.getString(R.string.flybits_con_default_tab_name), ignoreCase = true)){
                paramsBuilder.setLabels(category)
            }

            val params = paramsBuilder.build()
            contentResult = contentRepository.getContent(params,this)
        }
    }

    fun setFeedErrorDisplayer(feedErrorDisplayer: FeedErrorDisplayer) {
        this.feedErrorDisplayer = feedErrorDisplayer
    }

    override fun onSuccess(items: ArrayList<Content>) {
        Logger.d("Category: $category, onSuccess() items count: " + items.size)
        onRefreshFinishedCallback?.invoke()
        onRefreshFinishedCallback = null

        /*invalidate the database only if the local data has already been returned from the data
        * source otherwise you may interrupt the local data flow which will cause long loading*/
        loadedContent = true
        if (gotLocal){
            Logger.d("Category: $category, gotLocal = true, invalidating content data source")
            contentDataSource?.invalidate()
        }
    }

    override fun onException(exception: FlybitsException) {
        Logger.exception(javaClass.simpleName, exception)

        if (exception is NotConnectedException || exception is APIUsageExceededException) {
            flybitsConcierge.unauthenticateWithoutLogout(exception)
        }
        onRefreshFinishedCallback?.invoke()
        onRefreshFinishedCallback = null

        if (feedErrorDisplayer != null) {
            feedErrorDisplayer!!.onError(resourceProvider.getString(R.string.flybits_con_generic_error))
        }

    }

    override fun onLoadedAllItems() {
        loadedAll = true
    }

    fun connectAndDownLoad(activity: Activity, metaDataId: String, callback: ObjectResultCallback<Content>?) {
        FlybitsManager.isConnected(activity, true, object : ConnectionResultCallback {
            override fun onConnected() {
                Content.getById(activity, metaDataId, callback)
            }

            override fun notConnected() {

            }

            override fun onException(exception: FlybitsException) {
                callback?.onException(exception)
            }
        })

    }

    //these scroll events need to be queued to prevent a concurrent modification exception
    fun onScroll(visibleItems: List<BaseTemplate>){
        scrollEventQueue.add(ScrollEvent(visibleItems))
    }

    private fun setupScrollHandler(){
        Executors.newSingleThreadExecutor().execute {
            while (true){
                val visibleItems = scrollEventQueue.take().visibleItems

                val visibleContentList = visibleItems.asSequence().map { it.content }.filter { it != null }.toList()

                //Cancel then remove tasks which disappeared from the screen after scrolling
                currentTaskList.keys.asSequence()
                        .filter {
                            //Filter out all the content that can be found in the visible list
                            visibleContentList.find { item -> item.id == it } == null
                        }.forEach {
                            //Cancel the timer for the view that disappeared
                            currentTaskList[it]?.cancel()
                            currentTaskList.remove(it)
                        }

                //Schedule the view actions for views which just appeared on the screen after scrolling
                visibleContentList.forEach {
                    //This timer will be canceled if this view disappears before the 250ms timer is done
                    if (currentTaskList[it.id] == null){
                        currentTaskList[it.id] = Timer().schedule(VIEWING_THRESHOLD){
                            contentAnalytics.trackViewed(it, System.currentTimeMillis())
                            currentTaskList.remove(it.id)
                        }
                    }
                }
            }
        }
    }
}
